#ifndef __MATROXFB_TI3026_H__
#define __MATROXFB_TI3026_H__

/* make checkconfig does not walk through whole include tree */
#include <linux/config.h>

#include "matroxfb_base.h"

#ifdef CONFIG_FB_MATROX_MILLENIUM
extern struct matrox_switch matrox_millennium;
#endif

#endif	/* __MATROXFB_TI3026_H__ */
